package datarecord;

import java.util.*;
import java.io.PrintStream;

/**
 * Objects of this class allocate DataRecords.  When they're returned,
 * it caches them for subsequent re-allocation.  This class exists to
 * try to mitigate the overheads of Java on the results.
 *
 * @version $Id: DataRecordManager.java,v 1.2 2010/01/05 08:23:09 zahorjan Exp $
 */

public class DataRecordManager {
  /**
   * To try to minimize Java effects, we cache DataRecords (to avoid
   * creating a lot, which might require garbage collecting a lot).
   * There is a list per record capacity.
   */
  private HashMap<Integer, LinkedList<DataRecord>> freeMap = null;


  /**
   * Class factory -- returns a DataRecord.
   */
  public DataRecord getNewRecord(int size) {
    if ( freeMap == null ) {
      // need to create the free lists, one per size
      freeMap = new HashMap<Integer, LinkedList<DataRecord>>();
    }

    DataRecord record = null;
    Integer mapKey = new Integer(size);
    synchronized(freeMap) {
      LinkedList<DataRecord> freeList = freeMap.get( mapKey );
      if ( freeList != null && !freeList.isEmpty() ) {
        record = freeList.removeFirst();
        record.reInit();
      }
    }

    if ( record == null) {
      record = new DataRecord(size);
      record.mapKey = mapKey;
    }

    return record;
  }

  /**
   * Called when done using a DataRecord.
   */
  public void freeRecord( DataRecord record ) {
    // this list must already exist, because you couldn't have
    // a DataRecord to free otherwise

    synchronized(freeMap) {
      LinkedList<DataRecord> freeList = freeMap.get( record.mapKey );
      if ( freeList == null ) {
        freeList = new LinkedList<DataRecord>();
        freeMap.put( record.mapKey, freeList);
      }
      freeList.add(record);
    }
  }
}