package datasource;

import java.io.*;
import java.util.*;
import java.util.concurrent.*;
import java.net.*;

import datarecord.*;

/**
 * DataSource: abstract base class for components that provide
 * data through read() calls by their client.
 * <p>
 * A DataSource contains a thread of its own.  When created, the thread
 * is started, and goes into a loop reading data from its source and 
 * putting into a buffer.  A client thread call to the read() method
 * takes data out of the buffer, if there is any, and blocks until there
 * is some if the buffer is currently empty.
 * 
 * Thre read() interface operates on bytes -- the client asks for some 
 * maximum number of bytes. However, reading data a byte at a time from
 * data sources (like files or network connections) is slow.  Instead,
 * the DataSource wants to read bigger chunks for its own source.
 *
 * For that reason, the internal data buffer kept by a DataSource is
 * an array of DataRecords.  Each DataRecord is a byte[], of some reasonable
 * size.  The read thread in the DataSource tries to fetch a full DataRecord
 * from its data source whenever it does a read.
 * 
 * @version $Id: DataSource.java,v 1.11 2010/01/05 19:02:15 zahorjan Exp $
 */

abstract public class DataSource {

  /**
   * The data buffer between the local, reader thread and the client
   * thread.  (Accessed by client via the read() method.)
   */
  private ArrayBlockingQueue<DataRecord> dataBuffer;

  /**
   * A semaphore to coordinate access to the dataBuffer.
   */
  private Semaphore bufferSemaphore;

  /**
   * Object that knows how to create/cache DataRecords.
   */
  private DataRecordManager recordManager;

  /**
   * Size of dataBuffer;
   */
  private static final int BUFFER_SIZE = 20;

  /**
   * Size of each DataRecord in the buffer.
   *
   */
  private static final int RECORD_SIZE = 1024;

  /**
   * The reader thread (fetches data from this object's source and puts it in dataBuffer).
   */
  private Thread reader;

  /**
   * We need to keep track of what state the input source is in.
   * <ul>
   * <li>STARTING: object created but not yet connected to its own input
   * <li>CONNECTED: object created and connected to its input
   * <li>CLOSE_WAIT: object's source is out of data, but there is still data in the data
   *     buffer available to the client via read()
   * <li>CLOSED: out of data.  Further read() invocations are errors.
   */
  public final static int STARTING = 0;
  public final static int CONNECTED = 1;
  public final static int CLOSED = 3;

  protected int state = STARTING;

  /**
   * The ReceiveHelper implemented here depends on this abstract method
   * being implemented by the instantiated class.  That method should
   * read from the DataSource's input source, returning the number
   * of bytes read or 0 if there are no more.
   */
  abstract protected int readFromSrc(byte[] buf, int len) throws IOException;

  /**
   * All derived classes must be capable of reinitializing themselves.
   */
  abstract public void reInit() throws IOException, FileNotFoundException;

  /** 
   * Helper class to implement reading thread.  Loops reading byte[]'s
   * and putting them in the data buffer.
   */
  private class ReceiveHelper extends Thread {

    public void run() {

      try {
        while ( state == CONNECTED ) {

          DataRecord record = recordManager.getNewRecord(RECORD_SIZE);
          int lengthRead = readFromSrc( record.data, RECORD_SIZE);

          // If there won't be any more data ever, readFromSrc() should have returned -1
          // So, we're eventually going to put a record with length < 0 into the buffer.
          // (That's how read() will know there's nothing left.)
          record.length = lengthRead;
          dataBuffer.put( record );
          bufferSemaphore.release();  // indicate a record is available

          if ( lengthRead <= 0 ) {
            // our source has indicated it won't ever provide data again
            break;
          }
        }
      }
      catch ( Exception e ) {
        // any of a large variety of IO related exceptions
        // might have occurred.
        // Just terminate this thread
        state = CLOSED;
        System.err.println( e.getMessage() );
        e.printStackTrace();
      }
    }
  }

  /** 
   * Creates a new DataSource.
   */
  protected DataSource() {
    // create buffers to hold incoming packets
    dataBuffer = new ArrayBlockingQueue<DataRecord>(BUFFER_SIZE);
    // indicate nothing in them at present
    bufferSemaphore = new Semaphore(0);

    // create the DataRecord manager
    recordManager = new DataRecordManager();
  }

  /**
   * Function to initialize (called by derived class constructor).
   */
  public void readyToRead() {
    dataBuffer.clear();
    state = CONNECTED;

    //-------------------------------------------------
    // Start the receiving thread.
    //-------------------------------------------------

    reader = this.new ReceiveHelper();
    reader.start();
  }

  /**
   * Reads up to maxLength bytes and returns them via argument buf, starting at position 0.
   *
   * This method is synchronized because it performs read/update/write on the head element of the list.
   * If the client were multithreaded (which it's not) that could lead to trouble.
   *
   * @returns The number of bytes actually read.  A value less than 0 indicates that
   * there is no more data to be read.
   */

  public int read(byte[] buf, int maxLength) throws IOException, InterruptedException {

    // all out of data
    if ( state == CLOSED ) return -1;

    if ( maxLength <= 0 )
      throw new IOException("read() called with length " + maxLength);

    // wait until we're sure there's something in the buffer
    bufferSemaphore.acquire();

    // Now get access to the first record in the buffer, leaving
    // it in the buffer.
    DataRecord record = dataBuffer.peek();

    if ( record.length < 0 ) {
      // there's no more data, ever.  We leave the record in the buffer in case
      // the client calls read() again.
      state = CLOSED;
      return -1;
    }

    int outLength = record.length < maxLength ? record.length : maxLength;
    System.arraycopy(record.data, record.offset, buf, 0, outLength);

    // update record state to reflect bytes just consumed
    record.offset += outLength;
    record.length -= outLength;

    // If we've consumed all the bytes, remove the record.
    if ( record.length <= 0 ) {
      record = dataBuffer.take();
      recordManager.freeRecord( record );
    } else {
      bufferSemaphore.release();  // because we've left the record in the buffer
    }

    return outLength;
  }
}

