package datasource;

import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.net.*;

/**
 * A NetSource is a source that reads a file (sequentially).
 * <p>To allow code that runs more than one trial in a single run,
 * the server changes its port each trial, by successive adding +1, -1, +1, ...
 * to the base port number passed as a constructor argument.
 * (If it doesn't change port number, the client could connect twice 
 * to it before it has finished the first request, which it's not prepared to 
 * do).
 *
 * @version $Id: NetSource.java,v 1.8 2010/01/06 01:49:31 zahorjan Exp $
 */

public class NetSource extends DataSource {
  private int portIncrement; // the +1, -1, +1, ...
  private int port;
  private ServerSocket serverSocket;
  private Socket clientSocket;
  private InputStream inStream;

  private boolean haveConnection = false;

  public NetSource(int baseport) {
    port = baseport;
    portIncrement = 1;
  }

  /**
   * (Re)Initialize this object
   */
  public void reInit() throws IOException {
    serverSocket = new ServerSocket(port);

    // You can comment this line for quieter operation.
    System.err.println( "Waiting on socket " + port );

    // we'll hang on this call until a client connection is made.
    // we get back a new Socket that can be used to read from the client.
    // (It could be used to write to the client as well, but we're
    // not doing that in this application.)
    clientSocket = serverSocket.accept();

    // at this point the client side has contacted us and is probably
    // sending data.
    inStream = clientSocket.getInputStream();

    // We're done with the server socket, so let's close it now.
    serverSocket.close();
    // prepare for next reInit()
    port += portIncrement;
    portIncrement *= -1;

    readyToRead();  // start reader thread in DataSource
  }

  /**
   * Called by reader thread in DataSource object.
   */
  protected int readFromSrc(byte[] buf, int len) throws IOException {
    int readLength = inStream.read(buf, 0, len);
    if ( readLength < 0 ) close();
    return readLength;
  }

  /**
   * Closes the socket, and updates the port number to prepare
   * for being reInit()'ed
   */
  private void close() throws IOException {
    inStream.close();
    inStream = null;
    clientSocket.close();
  }

}
