package netapp;

import datasource.*;
import datasink.*;
import util.*;

/**
 * A Client is what we'd think of as the typical program.  In 
 * this case, the program doesn't do very much -- it just reads
 * from a source and copies what it gets to a sink.  That is, it 
 * does more or less what the *nix command 'cat' does.
 * <p>
 * There is somewhat more machinery in the code you'll find here than
 * that requires, but you should ignore it.  It has to do with
 * reliably running repeated executions of the Client while trying
 * to minimize the impact of Java on the results.  The essence of
 * a Client is simply "while my source has data, read it and write
 * it to my sink."
 *
 * @version $Id: Client.java,v 1.15 2010/01/06 02:18:56 zahorjan Exp $
 */


public class Client {

  /**
   * Used to compute timing statistics only.
   */
  public long elapsedTime;
  /**
   * Used to compute timing statistics only.
   */
  public long bytesSent;


  private static final int BUFSIZE = 1024;
  private byte[] buf;

  /**
   * Used for statistics keeping.
   */
  StopWatch watch;

  /**
   * Constructor: creates long-lived instance variables.
   */

  public Client() {
    buf = new byte[BUFSIZE];
    watch = new StopWatch();
  }

  /**
   * Run the Client.  All it does is copy from its
   * src to its sink.
   */

  public void go(DataSource src, DataSink sink) {
    elapsedTime = 0;
    bytesSent = 0;
    watch.mark();
    try {
      int len;
      do {
        len = src.read(buf, BUFSIZE);
        if ( len > 0 ) {
          sink.write(buf, len);
          bytesSent += len;
        }
      } while ( len > 0 );

    } catch (Exception e) {
      System.err.println("Exception caught by Client:");
      System.err.println( e.getMessage() );
      e.printStackTrace();
    }

    sink.close();
    // we wait until the data has at least actually been sent
    elapsedTime = watch.elapsedTime();
  }

  /**
   * Helper function to print some statistics.
   */

  private static void printStats(int nTrials, long elapsedTime,
                                 long totalBytesSent, long totalElapsedTime ) {

      double trialsElapsedTime = ((double)elapsedTime)/(1000*1000*1000);
      double avgElapsedTime = ((double)totalElapsedTime/nTrials) / (1000*1000*1000);
      double avgBytesSent = ((double)totalBytesSent)/nTrials;

      System.out.println( "\nTotal trials: " + nTrials + " (" + trialsElapsedTime + " seconds)" );
      System.out.println( "Processed " + avgBytesSent  + " bytes in an average " + avgElapsedTime + " seconds." );
      System.out.println( "Goodput: " + 8*avgBytesSent/(avgElapsedTime * (1024*1024)) + " Mbps");
        
  }


  /**
   * main() handles the command line arguments, meaning it sets up the src and sink
   * for the Client object. It runs as many repetitions of the experiment as it can
   * in a roughly fixed amount of time.
   */

  public static void main(String[] args) {
    DataSource src = null;
    DataSink sink = null;

    boolean amServer = false;

    // Create source and sink objects to connect the Client to

    int argIndex = 0;

    if ( args[argIndex].equals("null") ) {
      int size = Integer.parseInt( args[argIndex+1] );
      src = new NullSource(size);
      argIndex += 2;
    }
    else if ( args[argIndex].equals("file") ) {
      src = new FileSource(args[argIndex+1]);
      argIndex += 2;
    }
    else if ( args[argIndex].equals("net") ) {
      src = new NetSource(Integer.parseInt(args[argIndex+1]));
      amServer = true;
      argIndex += 2;
    }
    else {
      System.err.println("Didn't recognize source specifier: " + args[argIndex]);
      System.exit(1);
    }

    if ( args[argIndex].equals("null") ) {
      sink = new NullSink();
      argIndex++;
    }
    else if ( args[argIndex].equals("file") ) {
      sink = new FileSink(args[argIndex+1]);
      argIndex += 2;
    }
    else if ( args[argIndex].equals("net") ) {
      // arguments are hostname / port
      sink = new NetSink( args[argIndex+1], Integer.parseInt(args[argIndex+2]) );
      argIndex += 3;
    }
    else {
      System.err.println("Didn't recognize sink specifier: " + args[argIndex]);
      System.exit(2);
    }
        
    // sit in a loop running trials.  If acting as a server (meaning the source
    // is a NetSource object), never terminate.  Otherwise, terminate after
    // an approximately fixed amount of time.
    try {

      long totalBytesSent = 0;
      long totalElapsedTime = 0;

      long trialTimeLimit = 10L*1000*1000*1000;  // 10 seconds, in nanoseconds
      StopWatch trialsWatch = new StopWatch();

      Client client = new Client();

      // we force the number of trials to be even for robustness reasons when connecting
      // to a remote server.  It's a pain to restart the server to assure it's back to
      // listening on its base port -- it'd be nicer if it could just keep running.
      // IF everything's working and we run an even number of client trials, then the
      // server should be back on its base port...

      int trial;
      //for ( trial=0; trialsWatch.elapsedTime() < trialTimeLimit  || trial%2==1 || amServer; trial++ ) {
      for ( trial=0; trial<1; trial++ ) {

        src.reInit();
        sink.reInit();

        client.go(src, sink);

        totalElapsedTime += client.elapsedTime;
        totalBytesSent += client.bytesSent;

        // A server never stops running, so we print the statistics for the single
        // last execution here.
        //if (amServer) {
        if (false) {
          printStats(1, trialsWatch.elapsedTime(), totalBytesSent,  totalElapsedTime);
          totalElapsedTime = 0;
          totalBytesSent = 0;
          trialsWatch.mark();
        }
      }
      printStats(trial, trialsWatch.elapsedTime(), totalBytesSent,  totalElapsedTime);
    }
    catch (Exception e) {
      System.err.println("Client.main() caught exception:");
      System.err.println( e.getMessage() );
      e.printStackTrace();
      // If something has gone horribly wrong, need to do an exit(), as
      // there may be threads lying around -- they need to be terminated for
      // the application as a whole to terminate, so just returning here isn't
      // enough.
      System.exit(1);
    }
  }
}